from .logger import logger
from .tools import convert_slice_path, retry, retry_async
from .exceptions import (
    DownloadException,
    ChunkUnsupportedException,
    NotSupportedProtocolException,
    ConnectionException,
    AuthException
)
from .work import WorkerFuture
from .config import (
    DEFAULT_HEADERS,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_SLICED_CHUNK_SIZE,
    DEFAULT_TIMEOUT,
    DEFAULT_MAX_RETRY,
    DEFAULT_RETRY_DELAY,
    DEFAULT_SLICED_FILE_SUFFIX,
)
from .core import Result
from .equilibrium import DynamicSemaphore, DynamicConcurrencyController


import os
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse, unquote

import paramiko
from paramiko.ssh_exception import SSHException

from yundownload.core.resources import Resources
from yundownload.network.base import BaseProtocolHandler
from yundownload.utils import retry
from yundownload.utils.core import Result
from yundownload.utils.exceptions import ConnectionException, AuthException
from yundownload.utils.config import DEFAULT_CHUNK_SIZE
from yundownload.utils.logger import logger


class SFTPProtocolHandler(BaseProtocolHandler):
    def __init__(self):
        super().__init__()
        self.transport: Optional[paramiko.Transport] = None
        self.sftp: Optional[paramiko.SFTPClient] = None
        self.support_resume = True

    @staticmethod
    def check_protocol(uri: str) -> bool:
        """检查是否支持 SFTP 协议"""
        return uri.lower().startswith("sftp://")

    def download(self, resources: "Resources"):
        return self._download(resources)

    def _download(self, resources: "Resources"):
        """实现 SFTP 断点续传下载"""
        uri = resources.uri
        local_path = resources.save_path

        parsed = urlparse(uri)
        host = parsed.hostname
        port = parsed.port or resources.sftp_port or 22
        username = parsed.username or "anonymous"
        password = parsed.password or ""
        remote_path = unquote(parsed.path)

        self._connect(host, port)
        self._login(username, password)

        logger.info(f"Login success to {uri}")

        file_stat = self.sftp.stat(remote_path)
        file_size = file_stat.st_size

        prepare_result = self._prepare_local_file(local_path, file_size)
        if prepare_result & Result.EXIST:
            return Result.EXIST

        start_pos = local_path.stat().st_size if local_path.exists() else 0

        with self.sftp.open(remote_path, 'rb') as remote_file:
            remote_file.seek(start_pos)

            with open(local_path, 'ab' if start_pos > 0 else 'wb') as local_file:
                while True:
                    data = remote_file.read(DEFAULT_CHUNK_SIZE)
                    if not data:
                        break

                    local_file.write(data)
                    self.current_size += len(data)

        if local_path.stat().st_size != file_size:
            raise IOError("File size mismatch after download")

        return Result.SUCCESS

    def close(self):
        """关闭连接"""
        if self.sftp:
            self.sftp.close()
        if self.transport:
            self.transport.close()

    def _connect(self, host: str, port: int):
        """建立 SFTP 连接"""
        try:
            self.transport = paramiko.Transport((host, port)) # noqa
            self.transport.connect()
        except (SSHException, TimeoutError) as e:
            raise ConnectionException(f"SFTP connection failed: {e}")

    def _login(self, username: str, password: str):
        """用户认证"""
        try:
            self.transport.auth_password(username, password)
            if not self.transport.is_authenticated():
                raise AuthException("SFTP authentication failed")

            self.sftp = paramiko.SFTPClient.from_transport(self.transport)
        except SSHException as e:
            raise AuthException(f"Authentication error: {e}")

    @staticmethod
    def _prepare_local_file(local_path: Path, remote_size: int) -> 'Result':
        """准备本地文件（处理断点续传）"""
        if local_path.exists():
            local_size = local_path.stat().st_size
            if 0 < remote_size < local_size:
                local_path.unlink()
            elif local_size == remote_size:
                return Result.EXIST
        else:
            local_path.parent.mkdir(parents=True, exist_ok=True)
        return Result.UNKNOWN



import os
import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from yundownload.utils.core import Environment

from yundownload.utils.tools import Interval
from yundownload.utils import Result
from yundownload.utils.logger import logger

if TYPE_CHECKING:
    from yundownload.core import Resources


class BaseProtocolHandler(ABC):

    def __init__(self):
        """
        Protocol processor base class
        """
        self.start_time = time.time()
        self.current_size = 0
        self._last_current_size = 0
        self._last_time = time.time()
        self._total_size = 0
        self._total = 0
        self._steps = 0
        self.resources = None
        self.timer = Interval(int(os.getenv(Environment.LOG_EVERY, 5)), self._print)

    def _print(self):
        logger.resource_p2s(self.resources, self.progress, self.speed)

    @property
    def progress(self) -> float:
        """
        Get the progress of your download

        If it cannot be determined based on the file size or step progress, 0 will be returned

        :return: Download progress
        """
        if self._total != 0:
            return self._steps / self._total
        if self._total_size != 0:
            return round(self.current_size / self._total_size, 2)
        return 0

    @property
    def speed(self) -> float:
        """
        Get the download speed

        :return: Download speed
        """
        if self.current_size == 0:
            return 0
        current_size = self.current_size - self._last_current_size
        self._last_current_size = self.current_size
        speed = current_size / (time.time() - self._last_time)
        self._last_time  = time.time()
        return speed

    @staticmethod
    @abstractmethod
    def check_protocol(uri: str) -> bool:
        """
        Check if the URI is supported by the current protocol

        :param uri: The URI to check
        :return: True if protocol is supported, false otherwise
        """
        pass

    def __call__(self, resources: 'Resources') -> 'Result':  # noqa
        """
        Invoke the download method

        :param resources: Resource object
        :return: Result object
        """
        logger.resource_start(resources)
        try:
            self.timer.start()
            self.resources = resources
            result = self.download(resources)
            if result.is_success():
                logger.resource_result(resources, result)
            elif result.is_exist():
                logger.resource_exist(resources)
        except Exception as e:
            result = Result.FAILURE
            logger.resource_error(resources, e)
        finally:
            self.timer.cancel()
            self._print()

        return result

    @abstractmethod
    def download(self, resources: 'Resources') -> 'Result':
        """
        Download resources
        """
        pass

    @abstractmethod
    def close(self):
        """
        Turn off the resource
        """
        pass

    def __del__(self):
        self.close()

    def __repr__(self):
        return f"<{self.__class__.__name__}>"



from concurrent.futures import ProcessPoolExecutor, Future
from typing import TYPE_CHECKING, Type

from ..utils.work import WorkerFuture
from ..network.base import BaseProtocolHandler
from ..network.ftp import FTPProtocolHandler
from ..network.http import HttpProtocolHandler
from ..network.m3u import M3U8ProtocolHandler
from ..network.sftp import SFTPProtocolHandler
from ..utils.core import Result
from ..utils.exceptions import NotSupportedProtocolException
from ..utils.logger import logger
from ..utils.tools import retry

if TYPE_CHECKING:
    from ..core import Resources


def _run(protocols: Type['BaseProtocolHandler'], resources: 'Resources') -> 'Result':
    """
    Run the download callback

    :param protocols: Protocol Matcher
    :param resources: Resource Object
    :return: Result
    """
    return retry(retry_count=resources.retry, retry_delay=resources.retry_delay)(protocols())(resources)


class DownloadProcessPoolExecutor(ProcessPoolExecutor):
    """
    Download the process pool
    """

    def __init__(self, max_workers: int = None, **kwargs):
        super().__init__(max_workers, **kwargs)

    def run_download(self, protocol: Type['BaseProtocolHandler'], resources: 'Resources') -> 'Future[Result]':
        """
        提交下载任务

        :param protocol: Protocol Matcher
        :param resources: Resource Object
        :return: A Future object that returns the result
        """
        return super().submit(_run, protocol, resources)


class Downloader:
    """
    Downloader
    """

    def __init__(self, max_workers: int = 1):
        """
        Downloader

        :param max_workers: Maximum number of processes
        """
        self._protocols: list = [M3U8ProtocolHandler, HttpProtocolHandler, FTPProtocolHandler, SFTPProtocolHandler]
        self._lock_protocol = None
        self._download_pool = DownloadProcessPoolExecutor(max_workers=max_workers)

    def submit(self, resources: 'Resources') -> 'WorkerFuture':
        """
        提交任务

        :param resources: Resource Object
        :return:
        """
        if self._lock_protocol:
            protocol = self._lock_protocol
        else:
            protocol = self._match_protocol(resources)
        resources.lock()
        return WorkerFuture(
            future=self._download_pool.run_download(protocol, resources),
            protocol=protocol,
            resources=resources
        )

    def lock_protocol(self, protocol: BaseProtocolHandler):
        """
        Lock the protocol

        :param protocol: Protocol Matcher
        :return:
        """
        if self._lock_protocol:
            raise RuntimeError("A protocol is already locked.")
        self._lock_protocol = protocol

    def _match_protocol(self, resources: 'Resources') -> Type['BaseProtocolHandler']:
        """
        Match the download protocol

        :param resources: Resource Object
        :return: Protocol Matcher
        """
        for protocol in self._protocols:
            if protocol.check_protocol(resources.uri):
                logger.info(f"Protocol {protocol.__name__} is supported for {resources.uri}")
                return protocol
        raise NotSupportedProtocolException(resources.uri)

    def add_protocol(self, protocol_handler: BaseProtocolHandler):
        """
        Add a custom protocol handler

        :param protocol_handler: Protocol Matcher
        """
        if isinstance(protocol_handler, BaseProtocolHandler):
            self._protocols.insert(0, protocol_handler)
        else:
            raise TypeError("protocol_handler must be a subclass of BaseProtocolHandler "
                            "and implement its required methods")

    def remove_protocol(self, protocol_handler: BaseProtocolHandler):
        """
        Remove the custom protocol handler

        :param protocol_handler: Protocol Matcher
        """
        if isinstance(protocol_handler, BaseProtocolHandler):
            self._protocols.remove(protocol_handler)
        else:
            raise TypeError("protocol_handler must be a subclass of BaseProtocolHandler "
                            "and implement its required methods")

    def close(self):
        self._download_pool.shutdown()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from ftplib import FTP, error_perm, error_reply
from pathlib import Path
from urllib.parse import urlparse, unquote

from yundownload.core.resources import Resources
from yundownload.network.base import BaseProtocolHandler
from yundownload.utils.core import Result
from yundownload.utils.exceptions import ConnectionException, AuthException
from yundownload.utils.logger import logger


class FTPProtocolHandler(BaseProtocolHandler):
    def __init__(self):
        super().__init__()
        self.ftp = None
        self.support_rest = False
        self.support_size = False

    @staticmethod
    def check_protocol(uri: str) -> bool:
        """检查是否支持FTP协议"""
        return uri.lower().startswith("ftp://")

    def download(self, resources: "Resources"):
        return self._download(resources)

    def _download(self, resources: "Resources"):
        """实现断点续传的流式下载"""
        uri = resources.uri
        local_path = resources.save_path

        parsed = urlparse(uri)
        host = parsed.hostname
        port = parsed.port or resources.ftp_port or 21
        username = parsed.username or "anonymous"
        password = parsed.password or "anonymous@"
        remote_path = unquote(parsed.path)

        self._connect(host, port, resources.ftp_timeout)
        self._login(username, password)

        logger.info(f"Login success to {uri}")

        file_size = self._get_remote_size(remote_path)

        prepare = self._prepare_local_file(local_path, file_size)
        self._total_size = file_size
        if prepare & Result.EXIST:
            return Result.EXIST

        with open(local_path, "ab" if self.support_rest else "wb") as f:
            self.current_size += local_path.stat().st_size
            start_pos = f.tell()

            if self.support_rest and start_pos > 0:
                logger.info(f"FTP download resuming from {uri}")
                self.ftp.voidcmd(f"REST {start_pos}")

            def write_chunk(data: bytes):
                f.write(data) # noqa
                self.current_size += len(data)

            logger.info(f"FTP download started from {uri}")
            resp = self.ftp.retrbinary(f"RETR {remote_path}", write_chunk, rest=start_pos)

            if not resp.startswith("226"):
                logger.error(f"The FTP transfer did not complete {uri}")
                return Result.FAILURE

            return Result.SUCCESS

    def close(self):
        """关闭连接"""
        if self.ftp:
            try:
                self.ftp.quit()
            except Exception:
                pass
            self.ftp = None

    def _connect(self, host: str, port: int, timeout: int):
        """建立FTP连接"""
        self.ftp = FTP()
        try:
            self.ftp.connect(host, port, timeout=timeout)
            self._detect_capabilities()
        except Exception as e:
            raise ConnectionException(f"FTP connection failed: {e}")

    def _login(self, username: str, password: str):
        """用户认证"""
        try:
            resp = self.ftp.login(username, password)
            if "230" not in resp:
                raise AuthException("FTP authentication failed")
        except error_perm as e:
            raise AuthException(f"Authentication error: {e}")

    def _detect_capabilities(self):
        """检测服务器能力"""
        try:
            # 检测REST支持
            self.ftp.voidcmd("REST 0")
            self.support_rest = True
        except error_perm:
            self.support_rest = False

        try:
            # 检测SIZE支持
            self.ftp.voidcmd("TYPE I")
            self.ftp.size("")
            self.support_size = True
        except error_perm:
            self.support_size = False

    def _get_remote_size(self, remote_path: str) -> int:
        """获取远程文件大小"""
        if self.support_size:
            try:
                return self.ftp.size(remote_path)
            except error_reply:
                pass
        return 0

    def _prepare_local_file(self, local_path: Path, remote_size: int) -> 'Result':
        """准备本地文件（处理断点续传）"""
        if local_path.exists():
            local_size = local_path.stat().st_size
            if 0 < remote_size < local_size:
                local_path.unlink()
            elif self.support_rest and local_size == remote_size:
                return Result.EXIST
        else:
            local_path.parent.mkdir(parents=True, exist_ok=True)
        return Result.UNKNOWN


class DownloadException(Exception):
    """
    Base class for all download exceptions.
    """
    pass


class ChunkUnsupportedException(DownloadException):
    """
    Raised when the download server does not support chunked transfer encoding.
    """
    def __init__(self, uri: str):
        super().__init__(f"Chunked transfer encoding is not supported for URI: {uri}")

class NotSupportedProtocolException(DownloadException):
    """
    Raised when the protocol is not supported.
    """
    def __init__(self, uri: str):
        super().__init__(f"Unable to find specified protocol handler for URI: {uri}")

class ConnectionException(DownloadException):
    """
    Raised when the connection to the server fails.
    """
    def __init__(self, uri: str):
        super().__init__(f"Connection to server failed for URI: {uri}")

class AuthException(DownloadException):
    """
    Raised when the authentication fails.
    """
    def __init__(self, uri: str):
        super().__init__(f"Authentication failed for URI: {uri}")


from enum import IntFlag


class Environment:
    LOG_EVERY = 'YUNDOWNLOAD_LOG_EVERY'
    DEFAULT_CHUNK_SIZE = 'YUNDOWNLOAD_DEFAULT_CHUNK_SIZE'
    DEFAULT_SLICED_CHUNK_SIZE = 'YUNDOWNLOAD_DEFAULT_SLICED_CHUNK_SIZE'
    DEFAULT_TIMEOUT = 'YUNDOWNLOAD_DEFAULT_TIMEOUT'
    DEFAULT_MAX_RETRY = 'YUNDOWNLOAD_DEFAULT_MAX_RETRY'
    DEFAULT_RETRY_DELAY = 'YUNDOWNLOAD_DEFAULT_RETRY_DELAY'


class Result(IntFlag):
    SUCCESS = 1
    FAILURE = 2
    EXIST = 4
    WAIT = 8
    UNKNOWN = 16

    def is_success(self) -> bool:
        return bool(self & Result.SUCCESS)

    def is_failure(self) -> bool:
        return bool(self & Result.FAILURE)

    def is_exist(self) -> bool:
        return bool(self & Result.EXIST)

    def is_wait(self) -> bool:
        return bool(self & Result.WAIT)

    def __str__(self) -> str:
        return self.name.lower()



import asyncio
import collections
import math
import statistics
import time

from ..utils.logger import logger


class DynamicConcurrencyController:
    """
    Dynamic concurrency control classes
    """

    def __init__(self, min_concurrency=2, max_concurrency=30, window_size=100):
        # 并发控制参数
        self.min_concurrency = min_concurrency
        self.max_concurrency = max_concurrency
        self.current_concurrency = min_concurrency
        self.last_adjustment = time.monotonic()

        # 指标采样窗口
        self.window_size = window_size
        self.response_times = collections.deque(maxlen=window_size)
        self.successes = collections.deque(maxlen=window_size)
        self.failures = collections.deque(maxlen=window_size)

        # 状态跟踪
        self.base_response_time = None
        self.last_throughput = 0
        self.stability_counter = 0

        # 平滑处理参数
        self.ema_alpha = 0.3
        self.ema_response = 0

        # 信号量
        self.semaphore = asyncio.Semaphore(min_concurrency)

    def record_result(self, response_time: None, success=True):
        """记录每次请求的结果"""
        if response_time:
            self.response_times.append(response_time)
        if success:
            self.successes.append(1)
        else:
            self.failures.append(1)

        # 动态校准基准响应时间
        self._calibrate_base_response_time()

    def calculate_concurrency(self):
        """计算新的并发数值"""
        if len(self.response_times) < 10:  # 冷启动阶段
            return self._linear_ramp_up()

        # 核心指标计算
        success_rate = self._calculate_success_rate()
        rt_factor = self._calculate_response_time_factor()
        load_factor = self._calculate_load_factor()

        # 动态调整公式
        new_concurrency = self.current_concurrency * (
                (success_rate ** 1.5) *
                (rt_factor ** 0.7) *
                (load_factor ** 0.5)
        )

        # 应用自适应策略
        new_concurrency = self._apply_adaptive_policies(new_concurrency)

        # 边界保护和阻尼处理
        new_concurrency = max(self.min_concurrency,
                              min(self.max_concurrency, new_concurrency))
        new_concurrency = self._dampen_adjustment(new_concurrency)

        self.current_concurrency = new_concurrency
        return round(new_concurrency)

    def _calibrate_base_response_time(self):
        """动态校准基准响应时间（使用statistics替代numpy）"""
        if len(self.response_times) > 20:
            recent_rt = list(self.response_times)[-20:]
            try:
                new_base = statistics.median(recent_rt)
            except statistics.StatisticsError:
                new_base = recent_rt[len(recent_rt) // 2] if recent_rt else 0

            if self.base_response_time is None:
                self.base_response_time = new_base
            else:
                # 平滑过渡校准
                self.base_response_time = 0.8 * self.base_response_time + 0.2 * new_base

    def _calculate_success_rate(self):
        """计算加权成功率"""
        total = len(self.successes) + len(self.failures)
        if total == 0:
            return 1.0
        recent_success = sum(list(self.successes)[-10:]) / 10 if len(self.successes) >= 10 else 1.0
        historical_success = sum(self.successes) / total
        return 0.7 * recent_success + 0.3 * historical_success

    def _calculate_response_time_factor(self):
        """计算响应时间影响因子"""
        # EMA平滑处理响应时间
        current_rt = statistics.mean(self.response_times) if self.response_times else 0
        self.ema_response = (self.ema_alpha * current_rt +
                             (1 - self.ema_alpha) * self.ema_response)

        if self.base_response_time is None or self.base_response_time < 1e-5:
            return 1.0

        # 非线性响应关系
        rt_ratio = self.ema_response / self.base_response_time
        return 1.0 / (1.0 + math.log(max(1.0, rt_ratio)))

    def _calculate_load_factor(self):
        """计算系统负载因子"""
        throughput = len(self.successes) / (time.monotonic() - self.last_adjustment + 1e-7)
        throughput_ratio = throughput / (self.last_throughput + 1e-7)
        self.last_throughput = throughput

        return max(0.8, min(1.2, throughput_ratio))

    def _apply_adaptive_policies(self, concurrency):
        """应用自适应控制策略"""
        # 快速失败保护
        if len(self.failures) > len(self.successes):
            return concurrency * 0.5

        # 响应时间突增保护
        if len(self.response_times) > 30:
            recent_rt = statistics.mean(list(self.response_times)[-10:])
            if recent_rt > 3 * self.base_response_time:
                return concurrency * 0.7

        # 稳定性奖励
        if abs(concurrency - self.current_concurrency) < 0.1 * self.current_concurrency:
            self.stability_counter += 1
            if self.stability_counter > 5:
                return min(concurrency * 1.1, self.max_concurrency)
        else:
            self.stability_counter = 0

        return concurrency

    def _dampen_adjustment(self, new_concurrency):
        """调整阻尼系数避免震荡"""
        delta = new_concurrency - self.current_concurrency
        max_delta = 0.3 * self.current_concurrency  # 单次最大变化30%
        dampened_delta = max(-max_delta, min(delta, max_delta))
        return self.current_concurrency + dampened_delta

    def _linear_ramp_up(self):
        """冷启动阶段线性增长"""
        if time.monotonic() - self.last_adjustment > 5.0:  # 每5秒增长一次
            self.current_concurrency = min(self.current_concurrency + 1, self.max_concurrency)
            self.last_adjustment = time.monotonic()
        return self.current_concurrency

    def get_current_concurrency(self):
        """获取当前并发数"""
        return round(self.current_concurrency)

    def __len__(self):
        return self.calculate_concurrency()


class DynamicSemaphore(asyncio.Semaphore):
    """
    Dynamic control of concurrent semaphores
    """

    def __init__(self, dcc: 'DynamicConcurrencyController'):
        initial_permits = len(dcc)
        self._dcc = dcc
        super().__init__(value=initial_permits)
        self._target = initial_permits  # 当前目标并发数
        self._lock = asyncio.Lock()  # 状态修改锁

    async def adaptive_update(self):
        """动态调整信号量容量（线程安全）"""
        new_target = self._dcc.calculate_concurrency()
        if new_target == self._target:
            return
        await self.update(new_target)

    async def update(self, new_target: int):
        """动态调整信号量容量（线程安全）"""
        async with self._lock:
            logger.info(f"❤️Adjusting concurrency from {self._target} to {new_target}")
            if new_target < 0:
                raise ValueError("Concurrency cannot be negative")

            delta = new_target - self._target
            self._target = new_target

            # 调整可用许可数量
            if delta > 0:
                # 扩容：增加可用许可
                self._value += delta

                waiters = getattr(self, '_waiters', None)
                if waiters is not None:
                    # 确定waiters的长度，考虑它可能是列表或集合
                    waiters_count = len(waiters) if waiters else 0

                    # 唤醒等待者（最多唤醒delta个）
                    for _ in range(min(delta, waiters_count)):
                        self._wake_up_next()
            elif delta < 0:
                # 缩容：减少可用许可（不影响已获取许可）
                self._value = max(0, self._value + delta)

    @property
    def current_target(self):
        """获取当前目标并发数"""
        return self._target

    @property
    def available_permits(self):
        """获取当前可用许可数"""
        return self._value

    def record_result(self, rt: float = None, success: bool = True):
        self._dcc.record_result(rt, success)



import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
import httpx

from yundownload.network.base import BaseProtocolHandler
from yundownload.utils.config import DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE
from yundownload.utils.core import Result
from yundownload.utils.logger import logger
from yundownload.utils.tools import convert_slice_path

if TYPE_CHECKING:
    from yundownload.core.resources import Resources
    from yundownload.utils.equilibrium import DynamicSemaphore


class HttpProtocolHandler(BaseProtocolHandler):

    def __init__(self):
        super().__init__()
        self.client: None | httpx.Client = None
        self.aclient: None | httpx.AsyncClient = None
        self._method = 'GET'
        self._slice_threshold = None
        self.sliced_chunk_size = None

    def download(self, resources: 'Resources'):
        self._slice_threshold = resources.http_slice_threshold
        self._method = resources.http_method
        self.sliced_chunk_size = resources.http_sliced_chunk_size
        self.client = httpx.Client(
            timeout=resources.http_timeout,
            auth=resources.http_auth,
            mounts={
                'http://': httpx.HTTPTransport(
                    proxy=resources.http_proxy.get('http'),
                ),
                'https://': httpx.HTTPTransport(
                    proxy=resources.http_proxy.get('https'),
                )
            },
            headers=DEFAULT_HEADERS,
            follow_redirects=True,
            transport=httpx.HTTPTransport(
                retries=5
            ),
            verify=resources.http_verify
        )
        self.client.params.merge(resources.http_params)
        self.client.headers.update(resources.http_headers)
        self.client.cookies.update(resources.http_cookies)

        self.aclient = httpx.AsyncClient(
            timeout=resources.http_timeout,
            auth=resources.http_auth,
            mounts={
                'http://': httpx.AsyncHTTPTransport(
                    proxy=resources.http_proxy.get('http'),
                ),
                'https://': httpx.AsyncHTTPTransport(
                    proxy=resources.http_proxy.get('https'),
                )
            },
            headers=DEFAULT_HEADERS,
            follow_redirects=True,
            transport=httpx.AsyncHTTPTransport(
                retries=5
            ),
            verify=resources.http_verify
        )
        return self._match_method(resources)

    @staticmethod
    def check_protocol(uri: str) -> bool:
        check_uri = uri.lower()
        return check_uri.startswith('http') or check_uri.startswith('https')

    def _match_method(self, resources: 'Resources') -> Result:
        try:
            test_response = self.client.head(resources.uri)
            test_response.raise_for_status()
            content_length = int(test_response.headers.get('Content-Length', 0))
            self._total_size = content_length
        except httpx.HTTPStatusError:
            with self.client.stream(self._method, resources.uri, data=resources.http_data) as test_response:
                test_response.raise_for_status()
                content_length = int(test_response.headers.get('Content-Length'))
            test_response = self.client.request(self._method, resources.uri, headers={'Range': 'bytes=0-1'},
                                                data=resources.http_data)
        except Exception as e:
            logger.error(e, exc_info=True)
            return Result.FAILURE

        if resources.save_path.exists():
            if resources.save_path.stat().st_size == content_length:
                return Result.EXIST
            elif resources.save_path.stat().st_size > content_length:
                resources.save_path.unlink()
        resources.save_path.parent.mkdir(parents=True, exist_ok=True)
        breakpoint_flag = self._breakpoint_resumption(test_response)
        resources.metadata['_breakpoint_flag'] = breakpoint_flag
        if breakpoint_flag and content_length > self._slice_threshold:
            logger.info(f'sliced download: {content_length} {resources.uri} to {resources.save_path}')
            return asyncio.run(self._sliced_download(resources, content_length))
        else:
            logger.info(f'stream download: {resources.uri} to {resources.save_path}')
            return self._stream_download(resources, content_length)

    def _stream_download(self, resources: 'Resources', content_length: int) -> Result:
        headers = {}
        if resources.save_path.exists():
            file_size = resources.save_path.stat().st_size
            if file_size == content_length:
                logger.info(f'file exist skip download: {resources.uri} to {resources.save_path}')
                return Result.EXIST
            elif file_size > content_length:
                resources.save_path.unlink()
            else:
                headers['Range'] = f'bytes={file_size}-'

        with self.client.stream(self._method,
                                resources.uri,
                                headers=headers,
                                data=resources.http_data) as response:
            response.raise_for_status()
            if resources.metadata.get('_breakpoint_flag', False):
                file_mode = 'ab'
            else:
                file_mode = 'wb'
            with resources.save_path.open(file_mode) as f:
                self.current_size += resources.save_path.stat().st_size
                for chunk in response.iter_bytes(chunk_size=DEFAULT_CHUNK_SIZE):
                    f.write(chunk)
                    self.current_size += len(chunk)
        return Result.SUCCESS

    async def _sliced_download(self, resources: 'Resources', content_length: int) -> Result:
        path_template = convert_slice_path(resources.save_path)
        chunks_path = []
        tasks = []
        for start in range(0, content_length, self.sliced_chunk_size):
            end = start + self.sliced_chunk_size - 1
            if end > content_length:
                end = content_length
            slice_path = path_template(start)
            tasks.append(
                asyncio.create_task(
                    self._sliced_chunked_download(resources, slice_path, start, end, resources.semaphore)
                )
            )
            chunks_path.append(slice_path)
        results = await asyncio.gather(*tasks)
        await self.aclient.aclose()
        if all(results):
            self._merge_chunk(resources.save_path, chunks_path)
            return Result.SUCCESS
        return Result.FAILURE

    async def _sliced_chunked_download(self, resources: 'Resources', save_path: Path, start: int, end: int,
                                       sem: 'DynamicSemaphore') -> bool:
        async with sem:
            logger.info(f'start sliced download: {resources.uri} to {resources.save_path} {start}-{end}')
            headers = {'Range': f'bytes={start}-{end}'}
            if save_path.exists():
                chunk_file_size = save_path.stat().st_size
                if chunk_file_size == self.sliced_chunk_size:
                    logger.info(f'slice exist skip download: {resources.uri} to {save_path}')
                    self.current_size += save_path.stat().st_size
                    return True
                elif chunk_file_size > self.sliced_chunk_size:
                    logger.info(f'slice size is larger than the slice size: {resources.uri} to {save_path}')
                    save_path.unlink()
                elif chunk_file_size == end - start + 1:
                    logger.info(f'slice exist skip download: {resources.uri} to {save_path}')
                    self.current_size += save_path.stat().st_size
                    return True
                else:
                    file_start = start + chunk_file_size
                    logger.info(f'slice breakpoint resumption: {resources.uri} to {save_path}')
                    headers['Range'] = f'bytes={file_start}-{end}'
                    if file_start == end:
                        logger.info(f'slice exist skip download: {resources.uri} to {save_path}')
                        self.current_size += save_path.stat().st_size
                        return True
            async with self.aclient.stream(self._method,
                                           resources.uri,
                                           headers=headers,
                                           data=resources.http_data) as response:
                response: httpx.Response
                if not response.is_success: sem.record_result(success=False)
                response.raise_for_status()
                async with aiofiles.open(save_path, 'ab') as f:
                    self.current_size += save_path.stat().st_size
                    async for chunk in response.aiter_bytes(chunk_size=DEFAULT_CHUNK_SIZE):
                        await f.write(chunk)
                        self.current_size += len(chunk)
                sem.record_result(response.elapsed.total_seconds(), True)
                await sem.adaptive_update()
            logger.info(f'sliced download success: {resources.uri} to {save_path}')
            return True

    @staticmethod
    def _merge_chunk(save_path: Path, result: list[Path]):
        with save_path.open('wb') as f:
            for chunk_path in result:
                with chunk_path.open('rb') as f_chunk:
                    while True:
                        chunk = f_chunk.read(DEFAULT_CHUNK_SIZE)
                        if not chunk:
                            break
                        f.write(chunk)
                logger.info(f'merge chunk success: {save_path} to {chunk_path}')
        for chunk_path in result:
            chunk_path.unlink()
            logger.info(f'delete chunk success: {chunk_path}')
        logger.info(f'merge file success: {save_path}')

    @staticmethod
    def _breakpoint_resumption(response):
        return (response.headers.get('Accept-Ranges') == 'bytes' or
                response.headers.get('Content-Range', '').startswith('bytes 0-1/') or
                response.headers.get('Content-Length') == '2')

    def close(self):
        self.client.close()



from yundownload.network.base import BaseProtocolHandler
from yundownload.network.http import HttpProtocolHandler
from yundownload.network.ftp import FTPProtocolHandler
from yundownload.network.sftp import SFTPProtocolHandler
from yundownload.network.m3u import M3U8ProtocolHandler


import asyncio
import time
from pathlib import Path
from random import randint
from string import Template
from threading import Thread, Event
from typing import Callable, Union, TypeVar

from ..utils.config import DEFAULT_SLICED_FILE_SUFFIX
from ..utils.logger import logger

T = TypeVar('T')


def convert_slice_path(path: Path) -> Callable[[int], Path]:
    template_path = Template("{}--$slice_id{}".format(
        path.with_name(path.name.replace('.', '-')).absolute(),
        DEFAULT_SLICED_FILE_SUFFIX
    ))

    def render_slice_path(slice_id: int) -> Path:
        return Path(template_path.substitute(slice_id=slice_id))

    return render_slice_path


def retry(
        retry_count: int = 1,
        retry_delay: Union[int, tuple[float, float]] = 2
):
    """
    Retry the decorator

    :param retry_count: Number of retries
    :param retry_delay: Retry interval
    :return:
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        def wrapper(*args, **kwargs) -> T:
            for i in range(retry_count):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if i == retry_count - 1:
                        logger.error(f"Retry {i + 1}/{retry_count} times, error: {e}", exc_info=True)
                        raise e
                    logger.warning(f"Retry {i + 1}/{retry_count} times, error: {e}", exc_info=True)
                    if isinstance(retry_delay, tuple):
                        time.sleep(randint(*retry_delay))
                    else:
                        time.sleep(retry_delay)
            raise RuntimeError("Unreachable code")

        return wrapper

    return decorator


def retry_async(retry_count: int = 1, retry_delay: Union[int, tuple[float, float]] = 2):
    """
    Asynchronous retryer

    :param retry_count: Number of retries
    :param retry_delay: Retry interval
    :return:
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        async def wrapper(*args, **kwargs) -> T:
            for i in range(retry_count):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if i == retry_count - 1:
                        logger.error(f"Retry Async {i + 1}/{retry_count} times, error: {e}", exc_info=True)
                        raise e
                    logger.warning(f"Retry Async {i + 1}/{retry_count} times, error: {e}", exc_info=True)
                    if isinstance(retry_delay, tuple):
                        await asyncio.sleep(randint(*retry_delay))
                    else:
                        await asyncio.sleep(retry_delay)
            raise RuntimeError("Unreachable code")

        return wrapper

    return decorator


class Interval(Thread):
    """Call a function after a specified number of seconds:

            t = Timer(30.0, f, args=None, kwargs=None)
            t.start()
            t.cancel()     # stop the timer's action if it's still waiting

    """

    def __init__(self, interval, function, args=None, kwargs=None):
        Thread.__init__(self)
        self.interval = interval
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.finished = Event()

    def cancel(self):
        """Stop the timer if it hasn't finished yet."""
        self.finished.set()

    def run(self):
        while not self.finished.is_set():
            self.finished.wait(self.interval)
            self.function(*self.args, **self.kwargs)



import asyncio
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urljoin

import aiofiles
import m3u8
from httpx import AsyncClient, Response, AsyncHTTPTransport

from yundownload.network.base import BaseProtocolHandler
from yundownload.utils.core import Result
from yundownload.utils.logger import logger

if TYPE_CHECKING:
    from yundownload.core.resources import Resources
    from yundownload.utils.equilibrium import DynamicSemaphore

from yundownload.utils import DEFAULT_CHUNK_SIZE


class M3U8ProtocolHandler(BaseProtocolHandler):
    @staticmethod
    def check_protocol(uri: str) -> bool:
        parse = urlparse(uri)
        return parse.scheme in {'http', 'https'} and parse.path.endswith('.m3u8')

    def download(self, resources: 'Resources') -> 'Result':
        return asyncio.run(self.download_segments(resources))

    async def download_segments(self, resources: 'Resources') -> 'Result':
        """
        Download the m3u8 playlist

        :param resources: Resource objects
        :return: Result
        """
        if resources.save_path.exists():
            return Result.EXIST
        async with AsyncClient(
                auth=resources.http_auth,
                timeout=resources.http_timeout,
                params=resources.http_params,
                headers=resources.http_headers,
                cookies=resources.http_cookies,
                mounts={
                    'http://': AsyncHTTPTransport(
                        proxy=resources.http_proxy.get('http'),
                    ),
                    'https://': AsyncHTTPTransport(
                        proxy=resources.http_proxy.get('https'),
                    )
                },
                verify=resources.http_verify
        ) as client:
            final_playlist = await self.handle_variant_playlist(client, resources)
            segments = self.parse_segments(final_playlist)
            tasks = []
            video_path = resources.save_path.parent / f"{resources.save_path.stem}"
            video_path.mkdir(parents=True, exist_ok=True)
            self._total = len(segments)
            segment_paths = []
            for index, seg in enumerate(segments):
                segment_path = video_path / f"{index}.ts"
                tasks.append(
                    asyncio.create_task(
                        self.download_segment(index, seg, segment_path, client, resources.semaphore)
                    )
                )
                segment_paths.append(segment_path)

            results = await asyncio.gather(*tasks)

            if all([r & (Result.SUCCESS | Result.EXIST) for r in results]):
                if not segments[0]['encryption']:
                    await self.merge_segments(segment_paths, resources.save_path)
                else:
                    logger.info("This is a encrypted m3u8, please decrypt it by yourself")
                return Result.SUCCESS
            return Result.FAILURE

    async def download_segment(self, index: int, seg: dict, save_path: Path, client: 'AsyncClient',
                               sem: 'DynamicSemaphore') -> 'Result':
        """
        Download the clip
        There is no encryption here, you can decrypt it by rewriting the method
        and getting the value via seg['encryption'].

        :param index: Slice index
        :param seg: Fragment information
        :param save_path: The path to save the clip
        :param client: Network connection pooling
        :param sem: Asynchronous semaphore
        :return:
        """
        async with sem:
            logger.info(f"Downloading fragments #{index} encryption {bool(seg['encryption'])} from {seg['uri']}")
            async with client.stream('GET', seg['uri']) as response:
                response: Response
                if not response.is_success: sem.record_result(success=False)
                response.raise_for_status()
                if save_path.exists() and response.headers.get('Content-Length') == str(save_path.stat().st_size):
                    sem.record_result(success=True)
                    await sem.adaptive_update()
                    return Result.EXIST
                async with aiofiles.open(save_path, "wb") as f:
                    async for chunk in response.aiter_bytes(chunk_size=DEFAULT_CHUNK_SIZE):
                        await f.write(chunk)
                        self.current_size += len(chunk)
                sem.record_result(response.elapsed.total_seconds(), True)
                await sem.adaptive_update()
            logger.info(f"Download fragments #{index} success from {seg['uri']}")
            self._steps += 1
            return Result.SUCCESS

    @staticmethod
    async def merge_segments(segment_paths: list[Path], save_path: Path):
        async with aiofiles.open(save_path, "wb") as f:
            for segment_path in segment_paths:
                async with aiofiles.open(segment_path, "rb") as segment_file:
                    chunk_size = DEFAULT_CHUNK_SIZE
                    while True:
                        chunk = await segment_file.read(chunk_size)
                        if not chunk:
                            break
                        await f.write(chunk)
                logger.info(f"Merge fragments #{segment_path} to {save_path}")
        for segment_path in segment_paths:
            segment_path.unlink()
            logger.info(f"Delete fragments #{segment_path}")
        segment_path.parent.unlink()
        logger.info(f"Merge fragments success to {save_path}")

    @staticmethod
    def parse_segments(playlist: m3u8.M3U8) -> list:
        """Parse TS fragment information"""
        segments = []

        for seg in playlist.segments:
            segment_info = {
                'duration': seg.duration,
                'uri': urljoin(playlist.base_uri, seg.uri),
                'encryption': None
            }

            if seg.key:
                segment_info['encryption'] = {
                    'method': seg.key.method,
                    'key_uri': urljoin(playlist.base_uri, seg.key.uri) if seg.key.uri else None,
                    'iv': seg.key.iv.hex() if seg.key.iv else None
                }

            segments.append(segment_info)
        return segments

    async def handle_variant_playlist(self, client: 'AsyncClient', resources: 'Resources') -> 'm3u8.M3U8':
        """Process the main playlist and select the sub-list with the highest bitrate"""
        playlist = await self.m3u8_load(client, resources.uri)
        if not playlist.is_variant:
            return playlist

        logger.info(f'm3u8 contains {len(playlist.playlists)} bitrate: {resources.uri} to {resources.save_path}')

        # Select the sub-playlist with the highest bandwidth
        best_playlist = max(
            playlist.playlists,
            key=lambda p: p.stream_info.bandwidth
        )

        logger.info(
            f'Selected sub-bitrate: {best_playlist.uri}, bindwidth: {best_playlist.stream_info.bandwidth}bps resolution: {best_playlist.stream_info.resolution} codecs: {best_playlist.stream_info.codecs}')

        # Load the child playlist
        sub_url = urljoin(playlist.base_uri, best_playlist.uri)
        return await self.m3u8_load(client, sub_url)

    @staticmethod
    async def m3u8_load(client: 'AsyncClient', uri: str) -> 'm3u8.M3U8':
        response = await client.get(uri)
        response: Response
        return m3u8.M3U8(response.text, base_uri=urljoin(str(response.url), "."))

    def close(self):
        pass



from concurrent.futures import Future
from typing import Type, TYPE_CHECKING
from .core import Result

if TYPE_CHECKING:
    from network import BaseProtocolHandler
    from ..core.resources import Resources


class WorkerFuture:
    def __init__(self, future: Future, protocol: Type['BaseProtocolHandler'], resources: 'Resources'):
        self._future = future
        self._protocol = protocol
        self.resources = resources

    def wait(self):
        self._future.result()

    @property
    def state(self) -> 'Result':
        return self._future.result()

    def finish(self):
        return bool(self._future.result() & (Result.EXIST | Result.SUCCESS))

    def done(self):
        """
        :return: 是否完成
        """
        return self._future.done()

    def cancel(self):
        """
        取消任务
        :return:
        """
        return self._future.cancel()

    def running(self):
        """
        任务是否正在运行
        :return:
        """
        return self._future.running()

    def cancelled(self):
        """
        任务是否被取消
        :return:
        """
        return self._future.cancelled()

    def __repr__(self):
        return f'<WorkerFuture {self._future}>'






from .. import Downloader, Resources
import argparse

def cli():
    parser = argparse.ArgumentParser(
        description="Yun Download"
    )
    parser.add_argument('uri', help="资源链接")
    parser.add_argument('save_path', help="保存路径")
    args = parser.parse_args()
    with Downloader() as dl:
        resources = Resources(
            uri=args.uri,
            save_path=args.save_path
        )
        result = dl.submit(resources).state
        if result.is_failure():
            print(f'file download failed: {args.uri}')
        else:
            print(f'file download success: {args.uri}')




import os
from ..utils.core import Environment

DEFAULT_HEADERS = {
    'User-Agent': 'Wget/1.12 (linux-gnu)',
    'Content-Encoding': 'identity',
    'Accept-Encoding': 'identity'
}
DEFAULT_CHUNK_SIZE = int(os.getenv(Environment.DEFAULT_CHUNK_SIZE, 1024 * 1024))
DEFAULT_SLICED_CHUNK_SIZE = int(os.getenv(Environment.DEFAULT_SLICED_CHUNK_SIZE, 100 * 1024 * 1024))
DEFAULT_TIMEOUT = int(os.getenv(Environment.DEFAULT_TIMEOUT, 60))
DEFAULT_MAX_RETRY = int(os.getenv(Environment.DEFAULT_MAX_RETRY, 3))
DEFAULT_RETRY_DELAY = int(os.getenv(Environment.DEFAULT_RETRY_DELAY, 3))
DEFAULT_SLICED_FILE_SUFFIX = '.ydstf'



import logging
from typing import TYPE_CHECKING
import colorlog

if TYPE_CHECKING:
    from ..core import Resources
    from ..utils import Result


class Logger(logging.Logger):
    def __init__(self):
        super().__init__('download', level=logging.INFO)
        formatter = colorlog.ColoredFormatter(
            '%(log_color)s%(levelname)s - %(name)s(%(process)s) - %(asctime)s - %(message)s%(reset)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bg_white',
            }
        )

        handler = logging.StreamHandler()
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        self.addHandler(handler)

    def resource_start(self, resources: 'Resources'):
        self.info(f'🚀Start downloading metadata: {resources.uri} to {resources.save_path}')

    def resource_result(self, resources: 'Resources', result: 'Result'):
        self.info(f'🏁Downloading result: {result} metadata: {resources.uri} to {resources.save_path}')

    def resource_error(self, resources: 'Resources', error: Exception):
        self.error(f'🏗Downloading error: {error} metadata: {resources.uri} to {resources.save_path}', exc_info=True)

    def resource_exist(self, resources: 'Resources'):
        self.info(f'📦Downloading exist: metadata: {resources.uri} to {resources.save_path}')

    def resource_log(self, resources: 'Resources', message: str, lever: int | str = logging.INFO):
        self.log(lever, f'❓Downloading message: {message} metadata: {resources.uri} to {resources.save_path}')

    def resource_p2s(self, resources: 'Resources', progress: float, speed: float):
        self.info(
            f'📊Downloading progress: {progress} speed: {round(speed / 1024 / 1024, 2)} MB/S '
            f'metadata: {resources.uri} to {resources.save_path}')


logger = Logger()



from pathlib import Path
from typing import Union, Literal, Dict

from yundownload.utils import DynamicConcurrencyController, DynamicSemaphore


class Resources:
    _set_lock = True

    def __init__(self,
                 uri: str,
                 save_path: Union[str | Path],
                 http_method: Literal['GET', 'POST', 'PUT', 'DELETE'] = 'GET',
                 http_params: dict = None,
                 http_headers: dict = None,
                 http_data: dict = None,
                 http_proxy: Dict[Literal['http', 'https'], str] = None,
                 http_cookies: dict = None,
                 http_timeout: int = 30,
                 http_auth: tuple[str, str] = None,
                 http_verify: bool = False,
                 http_slice_threshold: int = 2048 * 1024 * 1024,
                 http_sliced_chunk_size: int = 2048 * 1024 * 1024,
                 ftp_timeout: int = 30,
                 ftp_port: int = 21,
                 sftp_port: int = 22,
                 metadata: dict = None,
                 retry: int = 3,
                 retry_delay: int | tuple[int, int] = 10,
                 min_concurrency: int = 2,
                 max_concurrency: int = 30,
                 window_size: int = 100):
        """
        Resource Object

        :param uri: Resource path
        :param save_path: The path where the resource is saved
        :param http_method: HTTP protocol request method
        :param http_params: HTTP request parameters (Valid for M3U8 protocol)
        :param http_headers: HTTP request headers (Valid for M3U8 protocol)
        :param http_data: HTTP request data
        :param http_proxy: HTTP request proxy { 'http': 'http://xxx', 'https': 'https://xxx' }
        :param http_cookies: HTTP request cookie (Valid for M3U8 protocol)
        :param http_timeout: HTTP request timeout period (Valid for M3U8 protocol)
        :param http_auth: HTTP protocol authentication is requested (Valid for M3U8 protocol)
        :param http_slice_threshold: HTTP protocol sharding threshold
        :param http_sliced_chunk_size: HTTP protocol sharding chunk size
        :param ftp_timeout: FTP request timeout period
        :param ftp_port: FTP protocol request port
        :param sftp_port: SFTP request port
        :param metadata: Custom metadata (for adapting custom protocols)
        :param retry: Number of retries
        :param retry_delay: Retry interval
        :param min_concurrency: Adaptive concurrency minimum
        :param max_concurrency: Adaptive concurrency maximum
        :param window_size: Adaptive concurrency window size
        """
        self.uri = uri
        self.save_path = Path(save_path)

        self.retry = retry
        self.retry_delay = retry_delay
        dcc = DynamicConcurrencyController(min_concurrency, max_concurrency, window_size)
        self.semaphore = DynamicSemaphore(dcc)

        # http protocol and part m3u8 protocol
        self.http_method = http_method
        self.http_params = http_params
        self.http_headers = http_headers
        self.http_data = http_data
        self.http_proxy = http_proxy if http_proxy else dict()
        self.http_cookies = http_cookies
        self.http_timeout = http_timeout
        self.http_auth = http_auth
        self.http_verify = http_verify
        self.http_slice_threshold = http_slice_threshold
        self.http_sliced_chunk_size = http_sliced_chunk_size

        self.ftp_timeout = ftp_timeout
        self.ftp_port = ftp_port

        self.sftp_port = sftp_port

        self.metadata = metadata if metadata else {}

    def lock(self):
        """
        Lock the resource object
        """
        self._set_lock = False

    def __setattr__(self, key, value):
        if self._set_lock:
            return super().__setattr__(key, value)
        raise AttributeError(f'{self.__repr__()} it is locked and cannot be modified')

    def __repr__(self):
        return "<Resources {} to {}>".format(self.uri, self.save_path)
